<?php

namespace WPStaging\Pro\Automations\PluginsUpdater;

use WPStaging\Core\WPStaging;
use WPStaging\Framework\Rest\Rest;
use WPStaging\Pro\Automations\BackgroundProcessing\PreparePluginsAutoUpdater;
use WPStaging\Pro\Automations\PluginsUpdater\Traits\PluginsUpdaterTrait;
use WPStaging\Pro\WPStagingPro;

use function WPStaging\functions\debug_log;

class WordPressPluginsUpdater extends AbstractPluginsUpdater
{
    use PluginsUpdaterTrait;

    /**
     * @param $options
     * @return void
     */
    public function autoUpdatePlugins($options)
    {
        if (empty($options->isAutoUpdatePlugins) || empty($options->cloneDirectoryName)) {
            return;
        }

        $stagingUrl = trailingslashit(get_home_url()) . $options->cloneDirectoryName;
        $data = [
            'name'                => 'wp_plugins_auto_updater',
            'stagingUrl'          => $stagingUrl,
            'isAutoUpdatePlugins' => $options->isAutoUpdatePlugins,
            'isWpCliRequest'      => true,
            'isNetworkClone'      => $this->isNetworkClone($options)
        ];

        $jobId = WPStaging::make(PreparePluginsAutoUpdater::class)->prepare($data);
        if ($jobId instanceof \WP_Error) {
            throw new \RuntimeException('Failed to trigger Plugins Updates in background: ' . $jobId->get_error_message());
        } else {
            debug_log('Plugins Updates triggered in background with job ID: ' . $jobId);
        }
    }

    /**
     * @return void
     */
    public function registerRestRoutes()
    {
        register_rest_route(Rest::WPSTG_ROUTE_NAMESPACE_V1, '/wpstg_plugins_update', [
            'methods'             => 'POST',
            'callback'            => [$this, 'mayBeUpdatePlugins'],
            'permission_callback' => '__return_true',
        ]);

        register_rest_route(Rest::WPSTG_ROUTE_NAMESPACE_V1, '/wpstg_get_outdated_plugins', [
            'methods'             => 'POST',
            'callback'            => [$this, 'getOutdatedPlugins'],
            'permission_callback' => '__return_true',
        ]);

        register_rest_route(Rest::WPSTG_ROUTE_NAMESPACE_V1, '/wpstg_activate_updated_plugins', [
            'methods'             => 'POST',
            'callback'            => [$this, 'activateUpdatedPlugin'],
            'permission_callback' => '__return_true',
        ]);
    }

    /**
     * @param \WP_REST_Request $request
     * @return \WP_Error|\WP_HTTP_Response|\WP_REST_Response
     */
    public function activateUpdatedPlugin(\WP_REST_Request $request)
    {
        $params     = $request->get_json_params();
        $pluginFile = $params['plugin_file'] ?? '';
        if (empty($pluginFile)) {
            return $this->sendResponse("Missing plugin file");
        }

        if (!$this->siteInfo->isStagingSite()) {
            return $this->sendResponse("You are not authorized to perform this action on Production site.");
        }

        if (is_plugin_active($pluginFile)) {
            return $this->sendResponse("Plugin already active");
        }

        // string $pluginFile, string $redirect = '', bool $network_wide = false, bool $silent = true (no need to run activation hook)
        $result = activate_plugin($pluginFile, '', false, true);
        if (is_wp_error($result)) {
            return $this->sendResponse(sprintf("Error activating the plugin: %s", $result->get_error_message()));
        }

        return $this->sendResponse("Plugin activated successfully");
    }

    /**
     * @return \WP_Error|\WP_HTTP_Response|\WP_REST_Response
     */
    public function mayBeUpdatePlugins(\WP_REST_Request $request)
    {
        $params         = $request->get_json_params();
        $pluginFile     = $params['plugin_file'] ?? '';
        $wpPluginUpdate = $params['wp_plugin_update'] ?? false;
        $isNetworkSite  = $params['is_network_site'] ?? false;
        $latestVersion  = $params['latest_version'] ?? '';
        $oldVersion     = $params['old_version'] ?? '';

        if ($wpPluginUpdate && !$isNetworkSite) {
            return $this->refreshPluginUpdates();
        }

        if ($wpPluginUpdate && $isNetworkSite) {
            $sites = get_sites();
            foreach ($sites as $site) {
                switch_to_blog($site->blog_id);
                if (version_compare($this->getPluginVersion($pluginFile), $latestVersion, '=')) {
                    continue;
                }

                return $this->refreshPluginUpdates();
            }
        }

        if (empty($pluginFile)) {
            return $this->sendResponse("Missing plugin file");
        }

        if (!$this->siteInfo->isStagingSite()) {
            return $this->sendResponse("You are not authorized to perform this action on Production site.");
        }

        if (is_multisite() && $isNetworkSite) {
            return $this->updatePluginsMultisite($pluginFile, $latestVersion, $oldVersion);
        }

        return $this->updatePlugins($pluginFile);
    }

    /**
     * @return \WP_Error|\WP_HTTP_Response|\WP_REST_Response
     */
    public function getOutdatedPlugins()
    {
        if (class_exists('WP_Plugin_Dependencies')) {
            \WP_Plugin_Dependencies::initialize();
        }

        wp_update_plugins();
        $updatePlugins   = get_site_transient('update_plugins');
        $outdatedPlugins = [];

        foreach ($updatePlugins->response as $pluginFile => $pluginData) {
            $pluginSlug    = dirname($pluginFile);
            $pluginDetails = get_plugin_data(WP_PLUGIN_DIR . '/' . $pluginFile);

            $outdatedPlugins[$pluginSlug] = [
                'name'            => $pluginDetails['Name'],
                'file'            => $pluginFile,
                'current_version' => $pluginDetails['Version'],
                'new_version'     => $pluginData->new_version,
                'slug'            => $pluginData->slug,
                'is_active'       => is_plugin_active($pluginFile),
                'dependencies'    => $this->getPluginDependencies($pluginFile)
            ];
        }

        $outdatedPlugins = $this->sortPluginsByDependencies($outdatedPlugins);

        return $this->sendResponse("Outdated Plugins", $outdatedPlugins);
    }

    /**
     * @param string $url
     * @return array
     */
    public function getOutdatedPluginsFromStagingSite(string $url = ''): array
    {
        if (empty($url)) {
            return [];
        }

        $apiResponse =  $this->getOutdatedPluginsRequest($url);
        if (is_wp_error($apiResponse)) {
            return [];
        }

        $outdatedPlugins = json_decode(wp_remote_retrieve_body($apiResponse), true);
        if (empty($outdatedPlugins)) {
            return [];
        }

        return $outdatedPlugins['result'] ?? [];
    }

    /**
     * @param string $url
     * @param bool $wpCheckUpdates
     * @return mixed|string
     */
    public function checkAvailableUpdates(string $url, bool $wpCheckUpdates)
    {
        $apiResponse = $this->sendPluginsUpdateRequest($url, '', $wpCheckUpdates);
        if (is_wp_error($apiResponse)) {
            return sprintf("Something went wrong while checking for available updates. Error: %s", $apiResponse->get_error_message());
        }

        $responseBody = json_decode(wp_remote_retrieve_body($apiResponse), true);
        if (empty($responseBody)) {
            return "Something went wrong while checking for available updates.";
        }

        return $responseBody['message'] ?? "Something went wrong while checking for available updates.";
    }

    /**
     * @param string $url
     * @param string $pluginFile
     * @param string $latestVersion
     * @param string $oldVersion
     * @param bool $isNetworkSite
     * @return array
     */
    public function sendUpdateRequest(string $url, string $pluginFile, string $latestVersion, string $oldVersion, bool $isNetworkSite = false): array
    {
        $apiResponse  = $this->sendPluginsUpdateRequest($url, $pluginFile, false, $latestVersion, $oldVersion, $isNetworkSite);
        $errorMessage = esc_html__("Error updating plugin. Please try updating manually", "wp-staging");

        if (is_wp_error($apiResponse)) {
            $errorMessage = sprintf(esc_html__("Error updating plugin: %s", "wp-staging"), $apiResponse->get_error_message());
            return $this->formatResponse($pluginFile, '0', '0', 'error', $errorMessage);
        }

        $responseBody = json_decode(wp_remote_retrieve_body($apiResponse), true);
        if (empty($responseBody)) {
            return $this->formatResponse($pluginFile, '0', '0', 'error', $errorMessage);
        }

        return $responseBody['result'] ?? [];
    }

    /**
     * @param string $url
     * @param string $pluginFile
     * @return string
     */
    public function activatePlugin(string $url, string $pluginFile): string
    {
        $apiResponse = $this->sendPluginActivationRequest($url, $pluginFile);
        if (is_wp_error($apiResponse)) {
            return sprintf("Error activating plugin: %s", $apiResponse->get_error_message());
        }

        $responseBody = json_decode(wp_remote_retrieve_body($apiResponse), true);
        if (empty($responseBody)) {
            return "Error activating plugin";
        }

        return $responseBody['message'] ?? '';
    }

    /**
     * @param array $response
     * @param string $stagingURL
     * @param array $outdatedPlugins
     * @return void
     */
    public function sendNotification(array $response, string $stagingURL, array $outdatedPlugins)
    {
        $subject = esc_html__("WP Staging Automation Update Report", "wp-staging");
        $message = $this->prepareNotificationBody($response, $stagingURL, $outdatedPlugins);
        $this->sendEmailNotification($subject, $message);
        if (!$this->sendSlackNotification($subject, $message)) {
            $this->sendSlackNotification($subject, $this->prepareNotificationBody($response, $stagingURL, $outdatedPlugins, true));
        }
    }

    /**
     * @param $options
     * @return bool
     */
    private function isNetworkClone($options)
    {
        if (!is_multisite() || empty($options->networkClone)) {
            return false;
        }

        return WPStagingPro::isValidLicense();
    }
}
