<?php

namespace WPStaging\Pro\Backup\Storage\Storages\GoogleDrive;

use WPStaging\Framework\Queue\FinishedQueueException;
use WPStaging\Pro\Backup\Storage\RemoteDownloaderInterface;
use WPStaging\Pro\Backup\Dto\Job\JobCloudDownloadDataDto;
use WPStaging\Vendor\Google\Client as GoogleClient;
use WPStaging\Vendor\Google\Service\Drive as GoogleDriveService;
use WPStaging\Vendor\Psr\Log\LoggerInterface;
use WPStaging\Backup\Service\BackupsFinder;
use WPStaging\Core\WPStaging;
use WPStaging\Framework\Adapter\Directory;

class Downloader implements RemoteDownloaderInterface
{
    /** @var GoogleClient */
    private $client;

    /** @var JobCloudDownloadDataDto */
    private $jobDataDto;

    /** @var LoggerInterface */
    private $logger;

    /** @var GoogleDriveService */
    private $service;

    /** @var int */
    private $chunkSize;

    /** @var Auth */
    private $auth;

    /** @var bool|string */
    private $error;

    /** @var Directory */
    private $dirAdapter;

    public function __construct(Auth $auth, Directory $directory)
    {
        $this->error = false;
        $this->auth = $auth;

        if (!$this->auth->isGuzzleAvailable()) {
            $this->error = 'cURL extension is missing. Backup is still available locally.';
            return;
        }

        if (!$this->auth->isAuthenticated()) {
            $this->error = 'Google Drive is not authenticated. Backup is still available locally.';
            return;
        }

        $this->client     = $auth->setClientWithAuthToken();
        $this->dirAdapter = $directory;
    }

    public function getProviderName(): string
    {
        return $this->auth->getLabel();
    }

    /**
     * @param LoggerInterface $logger
     * @param JobCloudDownloadDataDto $jobDataDto
     * @param int $chunkSize = MB_IN_BYTES
     * @return void
     */
    public function setupDownload(LoggerInterface $logger, JobCloudDownloadDataDto $jobDataDto, int $chunkSize = MB_IN_BYTES)
    {
        $this->logger     = $logger;
        $this->jobDataDto = $jobDataDto;
        $this->chunkSize  = $chunkSize < 10 * MB_IN_BYTES ? 10 * MB_IN_BYTES : $chunkSize;
    }

    /** @return string */
    public function getError()
    {
        return $this->error;
    }

    public function getBackups(): array
    {
        return $this->auth->getBackups();
    }

    /**
     * Delete backup from google drive
     *
     * @param string $fileId
     * @return bool
     */
    public function deleteFile(string $fileId): bool
    {
        $this->service = new GoogleDriveService($this->client);
        try {
            $backupFileName = $this->service->files->get($fileId)->getName();
            if ($this->service->files->delete($fileId)) {
                $this->auth->unsetStorageFromRetainedBackups($backupFileName);
                return true;
            }
        } catch (\Throwable $th) {
        }

        return false;
    }

    /**
     * Download backup file from google drive to server
     *
     * @param string $fileId
     * @param int $fileSize
     * @param int $chunkStart
     * @return int
     */
    public function chunkDownloadCloudFileToFolder(string $fileId, int $fileSize, int $chunkStart): int
    {
        $chunkSize           = $this->chunkSize;
        $downloadedChunkSize = 0;

        $this->service = new GoogleDriveService($this->client);
        $tmpDirectory  = $this->dirAdapter->getDownloadsDirectory();
        $file          = $this->service->files->get($fileId);
        $fileName      = $file->getName();
        $filePath      = $tmpDirectory . $fileName;
        $fileHandle    = fopen($filePath, 'a+');

        $chunkStart = filesize($filePath);
        $chunkEnd   = $chunkStart + $chunkSize;

        if ($chunkStart < $fileSize) {
            $http = $this->client->authorize();
            // There is no Google Drive API to download a file in chunks. So we use Guzzle to download the file in chunks.
            // @todo do more research about this later. see https://developers.google.com/drive/api/guides/manage-downloads#partial_download
            $response = $http->request(
                'GET',
                sprintf('https://www.googleapis.com/drive/v3/files/%s', $fileId),
                [
                    'query' => ['alt' => 'media'],
                    'headers' => [
                        'Range' => sprintf('bytes=%s-%s', empty($chunkStart) ? 0 : $chunkStart, $chunkEnd)
                    ]
                ]
            );

            fwrite($fileHandle, $response->getBody()->getContents());
            $contentLength = $response->getHeaderLine('Content-Length');
            $downloadedChunkSize = intval($contentLength) > 0 ? $chunkEnd : $downloadedChunkSize;
        }

        if ($chunkStart >= $fileSize || $downloadedChunkSize >= $fileSize) {
            // close the file pointer
            fclose($fileHandle);
            $backupsDirectory = WPStaging::make(BackupsFinder::class)->getBackupsDirectory();
            $destination      = $backupsDirectory . $fileName;
            // move backup from tmp to backup folder
            rename($filePath, $destination);
            throw new FinishedQueueException($fileName);
        }

        return $downloadedChunkSize;
    }
}
